%% Info
% Demo program to demonstrate the use of the
% Vectorial Kernel Orthogonal Greedy Algorithm (VKOGA)
% to train a kernel model.
% The code validates the RBF shape par. and the regularizatio par.
clear all, 
close all
startup()


%% Path of various files
% Data file
dataFile = 'Data/DataExample.mat';
% Permutation file used to split the data into train/validation/test.
% If it exists, this permutation is used. Otherwise a random permutation
% is used and saved with this name
permFile = 'Random/DataPermutationExample.mat';
% File where to save the trained model. Leave empty or remove to avoid saving
modelFile = 'Models/ModelExample.mat';
% Print info during training
verboseFlag = 1;
% Plot results after training
plotFlag = 1;


%% Parameters for VKOGA training and validation
% Data parameters
dataPar.percTe = 0.1;            % fraction of data to use as test set
dataPar.normalization = 'none';  % 'none' or 'uniform' (scale to [0, 1] ^ d)
dataPar.k = 5;                   % k-fold cross validation


% Training parameters
trPar.rbf = getRbf('gauss');     % rbf kernel (see getRbf.m)
trPar.tolF = 1e-10;              % tolerance on the residual
trPar.tolP = 1e-20;              % tolerance on the power function
trPar.maxExpSize = 100;          % max number of selected points
trPar.gType = 'pgreedy';         % greedy selection rule (fgreedy, pgreedy, fpgreedy)


% Validation parameters
valPar.spaceType = 'log';        % 'log' or 'lin' spacing
valPar.minEp = 1e-1;             % min value of shape parameter
valPar.maxEp = 1e1;               % max value of shape parameter
valPar.nEp = 5;                 % number of shape parameters
valPar.minLambda = 1e-10;        % min value of regularization parameter
valPar.maxLambda = 1e-2;         % max value of regularization parameter
valPar.nLambda = 4;              % number of regularization parameters
valPar.tolFEp = trPar.tolF;
valPar.tolPEp = trPar.tolP;
valPar.maxExpSizeEp = trPar.maxExpSize;


%% Load data
% It assumes dataFile contains 'input' and 'output', where input is
% an Nxd matrix of N d-dimensional points and output is a Nxq matrix of
% N q-dimensional points.
% Any other way to produce X, f is fine, as long as they contain one point 
% per row.
data = load(dataFile);
X = data.input;
f = data.output;


%% Create dataset
% - normalize data
% - permute them
% - split into train/validation/test

% Load or create a permutation
if exist('permFile', 'var') && ~isempty(permFile) && exist(permFile, 'file')
    dataPar.perm = importdata(permFile);
else
    warning('Using random permutation')
    pp = randperm(size(X, 1));
    dataPar.perm = pp;
    save(permFile, 'pp')
end

% Create the actual dataset
dataset = setData(X, f, dataPar);
clear('X'), clear('f')


%% Train and save the model
fprintf('Training kernel model\n')
[modelPar, trainResults, valResults, testResults] = trainModelReg(dataset, valPar, trPar, verboseFlag);


if exist('modelFile', 'var') && ~isempty(modelFile)
    save( modelFile, 'modelPar')
end


%% Plot training/validation/test errors
if plotFlag
    % size of model
    n = size(modelPar.a, 1);
    
    % Training error
    figure(1)
    semilogy(1 : n, trainResults.fmax, '.-', 'linewidth', 1.5)
    xlabel('Kernel expansion size')
    ylabel('Training error')
    set(gca, 'fontsize', 14)
    grid on
    
    % Power function on training set
    figure(2)
    semilogy(1 : n, trainResults.pmax, '.-', 'linewidth', 1.5), hold on
    xlabel('Kernel expansion size')
    ylabel('Max of power function')
    set(gca, 'fontsize', 14)
    grid on
    
    % Validation error
    if length(valResults.Lambda) > 1 && length(valResults.Eps) > 1
    figure(3)
    [xx, yy] = meshgrid(valResults.Lambda, valResults.Eps);
    if strcmp(valPar.spaceType, 'log')
        plot(log10(valResults.lambda), log10(valResults.ep), 'ro'), hold on
        contour(log10(xx), log10(yy), reshape(log10(valResults.errVal), size(xx)), 20, 'linewidth', 1.5)
        hold off
    else
        plot(valResults.lambda, valResults.ep, 'ro'), hold on
        contour(xx, yy, valResults.errVal, 20, 'linewidth', 1.5)
        hold off
    end
    legend([], 'Selected parameters')
    xlabel('Regularization parameter')
    ylabel('Shape parameter')
    title('Validation error (log_{10})')
    set(gca, 'fontsize', 14)
    grid on
    end
    
    if dataPar.percTe > 0
        % Test error (max and mean)
        figure(4)
        semilogy(1 : n, testResults.maxErr, '.-', 1 : n, testResults.meanErr, '.-', 'linewidth', 1.5)
        xlabel('Kernel expansion size')
        ylabel('Test error')
        set(gca, 'fontsize', 14)
        legend('Max', 'Mean')
        grid on
    end
end


