function [a, indI, n, c, Cut, fmax, pmax, Vx] = newtonYYDataMF(Afun, f, tolF, tolP, maxIter, gType, diagonal)
% NEWTONYYDATAMF runs a matrix-free VKOGA.
%   [a, indI, n] = NEWTONYYDATAMF(Afun, f, tolF, tolP, maxIter, gType)
%   returns the indexes indI of the selected points, the coefficient
%   matrix a of the corresponding interpolant, and the number n of selected
%   points.
%   The points are selected from a training set Xtr, which is implicitly
%   defined via the kernel matrix function Afun, which accepts two indexes
%   vectors as input, e.g.,
%   Afun((1 : 10)', (2 : 20)')
%   and returns the corresponding submatrix of the kernel matrix of Xtr.
%   The interpolant can then be evaluated in the test points Xte as
%   ste = ker(ep, Xte, Xtr(indI, :)) * a;
%   where ker = ker(ep, x, y) is a SPD kernel.
%   The matrix f contains the evaluations of the unknown function on the
%   training points, one point per row.
%   The algorithm is terminated by tolerances tolF on the square of the
%   residual, tolP on the square of the power function, or when maxIter
%   points are selected.
%   The greedy selection is made by the criterion gType, which can be one
%   among fgreedy, pgreedy, fpgreedy.
%   The implementation is based on the Matrix-Free implementation of the
%   Newton basis by M. Pazouki and R. Schaback. This means that only the
%   rows/columns of the kernel matrix corresponding to the selected points
%   are computed, while the full kernel matrix is never stored nor
%   computed.


[N, q] = size(f);               % number of points and dimension of output
maxIter = min(maxIter, N);      % reduces the max number of points if > N
indI = zeros(maxIter, 1);       % selected indexes
notIndI = (1 : N)';             % not yet selected indexes
Vx = zeros(N, maxIter);         % values of the Newton basis on the points
c = zeros(maxIter, q);          % coefficients of the interpolant/expansion
fmax = zeros(maxIter, 1);       % maximal residuals on training set
pmax = zeros(maxIter, 1);       % maximal power function on training set
if ~exist('diagonal', 'var') || isempty(diagonal)
    p = Afun(1, 1) * ones(N, 1);    % power function squared (assuming transl. inv. ker.)
else
    p = diagonal;
end
Cut = zeros(maxIter);           % matrix of change of basis

% Iterative selection of new points
for n = 1 : maxIter
    % select the current index
    if strcmp(gType, 'fgreedy')
        [fmax(n), i] = max(sum(f(notIndI, :) .^ 2, 2));
        pmax(n) = max(p(notIndI));
    elseif strcmp(gType, 'fpgreedy')
        [~, i] = max(sum(f(notIndI, :) .^ 2, 2) ./ p(notIndI));
        fmax(n) = max(sum(f(notIndI, :) .^ 2, 2));
        pmax(n) = max(p(notIndI));
    elseif strcmp(gType, 'pgreedy')
        [pmax(n), i] = max(p(notIndI));
        fmax(n) = max(sum(f(notIndI, :) .^ 2, 2));
    else
        error('Not a valid greedy method')
    end
    % add the current index
    indI(n) = notIndI(i);
    
    if fmax(n) >= tolF && pmax(n) >= tolP
        % compute the nth basis
        Vx(notIndI, n) = Afun(notIndI, indI(n)) - Vx(notIndI, 1 : n - 1) * Vx(indI(n), 1 : n - 1)';
        % normalize the nth basis
        Vx(notIndI, n) = Vx(notIndI, n) / sqrt(p(indI(n)));
        % update the change of basis
        Cut(n, 1 : n) = [-Vx(indI(n), 1 : n - 1) * Cut(1 : n - 1, 1 : n - 1),  1] / Vx(indI(n), n);
        % compute the nth coefficient
        c(n, :) = f(indI(n), :) ./ sqrt(p(indI(n)));
        % update the power function
        p(notIndI) = p(notIndI) - Vx(notIndI, n) .^ 2;
        % update the residual
        f(notIndI, :) = f(notIndI, :) - bsxfun(@times, Vx(notIndI, n), c(n, :));
        % remove the nth index from the dictionary
        notIndI = notIndI([1 : i - 1, i + 1 : end]);
    else
        n = n - 1;
        break
    end
end

% Final resize to match the actual number of selected points
c = c(1 : n, :);
Cut = Cut(1 : n, 1 : n);
a = Cut' * c;
indI = indI(1 : n);
fmax = sqrt(fmax(1 : n));
pmax = sqrt(pmax(1 : n));

end