function [modelPar, trainResults, valResults, testResults] = trainModelReg(dataset, valPar, trPar, verboseFlag)
% TRAINMODELREG trains a kernel model using VKOGA, with optimization of the
% kernel shape parameter and the regularizatio parameter via k-fold cross
% validation.


% Kernel
ker = @(ep, x, y) trPar.rbf(ep, distanceMatrix(x, y));

% Parameter optimization
errFun = @(s, f) sqrt(sum((f - s) .^ 2, 2)); 
if ~isfield(valPar, 'spaceType') || strcmp(valPar.spaceType, 'lin')
    Eps = linspace(valPar.minEp, valPar.maxEp, valPar.nEp);
    Lambda = linspace(valPar.minLambda, valPar.maxLambda, valPar.nLambda);
elseif strcmp(valPar.spaceType, 'log')
    Eps = logspace(log10(valPar.minEp), log10(valPar.maxEp), valPar.nEp);
    Lambda = logspace(log10(valPar.minLambda), log10(valPar.maxLambda), valPar.nLambda);
end

[ee, ll] = ndgrid(Eps, Lambda);
parameters = [ee(:) ll(:)];
nPar = size(parameters, 1);

errVal = zeros(nPar, 1);
fprintf('Parameter validation:\n')

parfor k = 1 : nPar
        tmp_err = zeros(size(dataset.indVal, 1), 1);
        for l = 1 : size(dataset.indVal, 1)
            Xtr = dataset.Xtr(setdiff(1 : size(dataset.Xtr, 1), dataset.indVal(l, 1) : dataset.indVal(l, 2)), :);
            ftr = dataset.ftr(setdiff(1 : size(dataset.Xtr, 1), dataset.indVal(l, 1) : dataset.indVal(l, 2)), :);
            Xval = dataset.Xtr(dataset.indVal(l, 1) : dataset.indVal(l, 2), :);
            fval = dataset.ftr(dataset.indVal(l, 1) : dataset.indVal(l, 2), :);
            Afun = @(i, j) ker(parameters(k, 1), Xtr(i, :), Xtr(j, :)) + parameters(k, 2) * double(bsxfun(@eq, i(:), j(:)'));
            [a, indX, n] = newtonYYDataMF(Afun, ftr, valPar.tolFEp, valPar.tolPEp, valPar.maxExpSizeEp, trPar.gType);
            s = ker(parameters(k, 1), Xval, Xtr(indX, :)) * a;
            tmp_err(l) = max(errFun(s, fval));
        end
        errVal(k) = mean(tmp_err);
        if verboseFlag
            fprintf('ep = %2.2e, lambda = %2.2e, error = %2.2e, n = %2d\n', parameters(k, 1), parameters(k, 2), errVal(k), n)
        end
end

[~, minErrParameter] = min(errVal);
epOpt = parameters(minErrParameter, 1);
lambdaOpt = parameters(minErrParameter, 2);
fprintf('Selected paramers: ep = %2.2e, lambda = %2.2e\n', epOpt, lambdaOpt)

% Model training
Afun = @(i, j) ker(epOpt, dataset.Xtr(i, :), dataset.Xtr(j, :)) + lambdaOpt * double(bsxfun(@eq, i(:), j(:)'));
[a, indX, n, c, Cut, fmax, pmax] = newtonYYDataMF(Afun, dataset.ftr, trPar.tolF, trPar.tolP, trPar.maxExpSize, trPar.gType);

% Define model parameters
modelPar.ep = epOpt;
modelPar.rbf = func2str(trPar.rbf);
modelPar.translX = dataset.translXtr;
modelPar.scaleX = dataset.scaleXtr;
modelPar.translF = dataset.translFtr;
modelPar.scaleF = dataset.scaleFtr;

% Test error computation
fprintf('Test error computation:\n')
if size(dataset.Xte, 1)
    errFun = @(s) sqrt(sum((dataset.fte - s) .^ 2, 2)); % ./ sqrt(sum((dataset.fte) .^ 2, 2)); 
    
    meanErr = zeros(n, 1);
    maxErr = zeros(n, 1);
    for k = 1 : n
        ak = Cut(1 : k, 1 : k)' * c(1 : k, :);
        modelPar.X = dataset.Xtr(indX(1 : k), :);
        modelPar.a = ak;
        s = kernelModel(dataset.Xte, modelPar);
        pointwiseErr =  errFun(s);
        meanErr(k) = mean(pointwiseErr);
        maxErr(k) = max(pointwiseErr);
    end
else
    meanErr = [];
    maxErr = [];
end

% Update model parameters
modelPar.X = dataset.Xtr(indX, :);
modelPar.a = a;

% Set up outputs
trainResults.fmax = fmax;
trainResults.pmax = pmax;

valResults.ep = modelPar.ep;
valResults.lambda = lambdaOpt;
valResults.Eps = Eps;
valResults.Lambda = Lambda;
valResults.errVal = errVal;

testResults.meanErr = meanErr;
testResults.maxErr = maxErr;

end


